@@ -1,6 +1,6 @@
-/*BINFMTC: parameter.c
+/*BINFMTC: parameter.c forkexec.c ilistcreate.c main.c
  *  cowbuilder / pbuilder with cowdancer
- *  Copyright (C) 2005-2007 Junichi Uekawa
+ *  Copyright (C) 2005-2009 Junichi Uekawa
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
  *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
  *
  */
-/* 
+/*
   sudo ./cowbuilder --create --distribution sid --basepath /mnt/120g-1/dancer/cow/cow
   sudo ./cowbuilder --update --basepath /mnt/120g-1/dancer/cow/cow
 
@@ -38,85 +38,12 @@ will show the list of files changed.
 
  */
 
-/* 
-
-pbuilder update time measurement on iBook G4 1GHz:
-
-cowbuilder:
-real    0m16.792s
-user    0m4.395s
-sys     0m3.568s
-
-pbuilder:
-real    2m30.573s
-user    0m38.957s
-sys     0m6.523s
-
-pbuilder build time with network down (no package install) on iBook G4
-1GHz:
-
-cowbuilder:
-real    0m18.694s
-user    0m7.899s
-sys     0m2.872s
-
-pbuilder:
-real    1m20.407s
-user    0m9.688s
-sys     0m4.142s
-
-with network up:
-cowbuilder:
-real    1m26.027s
-user    0m21.582s
-sys     0m7.377s
-
-pbuilder:
-real    2m57.126s
-user    0m22.255s
-sys     0m7.961s
-
-
-pbuilder login
-cowbuilder:
-real    0m4.488s
-user    0m1.039s
-sys     0m1.393s
-
-pbuilder:
-real    1m20.058s
-user    0m4.526s
-sys     0m3.689s
-
-
-
-ibookg4 on batteries, building cowdancer sources.
-time pdebuild --pbuilder cowbuild
-real    0m41.501s
-user    0m22.473s
-sys     0m8.473s
-
-real    0m45.087s
-user    0m22.422s
-sys     0m8.434s
-
-time debuild
-real    0m51.879s
-user    0m18.079s
-sys     0m4.401s
-
-2nd try:
-real    0m24.262s
-user    0m18.113s
-sys     0m3.792s
-
- */
-
 #define _GNU_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <string.h>
 #include <unistd.h>
 #include "parameter.h"
@@ -124,6 +51,19 @@ sys     0m3.792s
 
 const char* ilist_PRGNAME="cowbuilder";
 
+
+/**
+   remove a directory.
+
+   be careful about bind-mounted directories.
+
+   @return 0 on success
+ */
+static int rmrf(const char* path)
+{
+  return forkexeclp("rm", "rm", "-rf", path, NULL);
+}
+
 /**
  * @return .ilist path as malloc()ed string, or NULL on error.
  */
@@ -139,7 +79,98 @@ static char* get_ilistfile_path(const struct pbuilderconfig* pc)
   return ilistfile;
 }
 
-/* 
+/**
+   break hardlink.
+
+   return 0 on success, 1 on fail.
+ */
+int break_cowlink(const char* s)
+{
+  char *backup_file;
+  struct stat buf;
+  int pid;
+  int status;
+
+  if(lstat(s, &buf))
+    return 1;			/* if stat fails, the file probably
+				   doesn't exist; return, success */
+  if (S_ISLNK(buf.st_mode))
+    return 1;			/* ignore symbollic link */
+
+  if((buf.st_nlink > 1) && 	/* it is hardlinked */
+     S_ISREG(buf.st_mode))
+    {
+      if (asprintf(&backup_file, "%sXXXXXX", s)==-1)
+	{
+	  ilist_outofmemory("out of memory in break_cowlink");
+	  return 1;
+	}
+      switch(pid=fork())
+	{
+	case 0:
+	  /* child process, run cp */
+	  putenv("COWDANCER_IGNORE=yes");
+	  execl("/bin/cp", "/bin/cp", "-a", s, backup_file, NULL);
+	  perror("execl:cp:");
+	  exit(EXIT_FAILURE);
+	case -1:
+	  /* error condition in fork(); something is really wrong */
+	  ilist_outofmemory("out of memory in check_inode_and_copy, 2");
+	  break;
+	default:
+	  /* parent process, waiting for cp -a to terminate */
+	  if(-1==waitpid(pid, &status, 0))
+	    {
+	      perror("waitpid:cp");
+	      fprintf(stderr, "%s: unexpected waitpid error when waiting for process %i with status %x\n",
+		      ilist_PRGNAME, pid, status);
+	      goto error_spm;
+	    }
+
+	  if (!WIFEXITED(status))
+	    {
+	      /* something unexpected */
+	      fprintf(stderr, "%s: unexpected WIFEXITED status in waitpid: %x\n", ilist_PRGNAME,
+		      (int)status);
+	      goto error_spm;
+	    }
+	  else if (WEXITSTATUS(status))
+	    {
+	      /* cp -a failed */
+	      fprintf(stderr, "%s: cp -a failed for %s\n", ilist_PRGNAME, backup_file);
+	      goto error_spm;
+	    }
+	  /* when cp -a succeeded, overwrite the target file from the temporary file with rename */
+	  else if (-1==rename(backup_file, s))
+	    {
+	      perror ("file overwrite with rename");
+	      fprintf(stderr, "while trying rename of %s to %s\n",  backup_file, s);
+	      goto error_spm;
+	    }
+	}
+    }
+  return 0;
+
+ error_spm:
+  free(backup_file);
+  return 1;
+}
+
+/**
+   Convenience function to call break_cowlink with basepath prepended.
+ */
+static void break_cowlink_inside_chroot(const char* basepath, const char* filepath)
+{
+  char* filename;
+  asprintf(&filename, "%s%s", basepath, filepath);
+  break_cowlink(filename);
+  free(filename);
+}
+
+/**
+   hardlink-copy directory for COW use.  Un-hardlink some files which
+   do not work with hardlinking.
+
    @return 0 on success, 1 on failure.
  */
 static int cpbuilder_internal_cowcopy(const struct pbuilderconfig* pc)
@@ -147,12 +178,13 @@ static int cpbuilder_internal_cowcopy(const struct pbuilderconfig* pc)
   char *ilistfile;
 
   printf(" -> Copying COW directory\n");
-  if (0!=forkexeclp("rm", "rm", "-rvf", pc->buildplace, NULL))
+  if (0!=rmrf(pc->buildplace))
     return 1;
   if (0!=forkexeclp("cp", "cp", "-al", pc->basepath, pc->buildplace, NULL))
     return 1;
 
-  /* delete existing ilist file if it exists, because I use COWDANCER_REUSE */
+  /* ilist file handling.
+     Delete existing ilist file if it exists, because I use COWDANCER_REUSE */
   if(!(ilistfile=get_ilistfile_path(pc)))
     {
       /* outofmemory */
@@ -164,19 +196,32 @@ static int cpbuilder_internal_cowcopy(const struct pbuilderconfig* pc)
       /* if there was no ilist file in the beginning, that's not a
 	 problem.
        */
+      fprintf(stderr, "I: unlink for ilistfile %s failed, it didn't exist?\n",
+	      ilistfile);
+    }
+  else
+    {
+      fprintf(stderr, "I: removed stale ilistfile %s\n",
+	      ilistfile);
     }
   free(ilistfile);
-  
+
+  /* Non-hardlinkable file support
+   */
+  break_cowlink_inside_chroot(pc->buildplace, "/var/cache/debconf/config.dat");
+  break_cowlink_inside_chroot(pc->buildplace, "/var/cache/debconf/templates.dat");
+  break_cowlink_inside_chroot(pc->buildplace, "/var/cache/debconf/passwords.dat");
+
   return 0;
 }
 
-/* 
+/*
    @return 0 on success, 1 on failure.
  */
 static int cpbuilder_internal_cleancow(const struct pbuilderconfig* pc)
 {
   printf(" -> Cleaning COW directory\n");
-  if (0!=forkexeclp("rm", "rm", "-rf", pc->buildplace, NULL))
+  if (0!=rmrf(pc->buildplace))
     return 1;
   return 0;
 }
@@ -204,14 +249,14 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile_)
       return 1;
     }
 
-  
+
   if (0>asprintf(&buf_chroot, "chroot %s cow-shell", pc->buildplace))
     {
       /* outofmemory */
       fprintf(stderr, "cowdancer: out of memory.\n");
       return -1;
     }
-  
+
   if(!(ilistfile=get_ilistfile_path(pc)))
     {
       /* outofmemory */
@@ -222,18 +267,43 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile_)
   prevdir=get_current_dir_name();
   chdir(pc->buildplace);
 
-  if (forkexeclp("chroot", 
+  if (forkexeclp("chroot",
 		 "chroot",
 		 pc->buildplace,
-		 "cowdancer-ilistcreate", 
-		 "/.ilist", 
-		 "find . -xdev -path ./home -prune -o \\( \\( -type l -o -type f \\) -a -links +1 -print0 \\) | xargs -0 stat --format '%d %i '", 
+		 "cowdancer-ilistcreate",
+		 "/.ilist",
+		 "find . -xdev -path ./home -prune -o \\( \\( -type l -o -type f \\) -a -links +1 -print0 \\) | xargs -0 stat --format '%d %i '",
 		 NULL))
     {
       /* if there was an error, back off to manual form */
-      fprintf(stderr, "W: cowdancer-ilistcreate failed to run within chroot, falling back to old method\n");
-      ilistcreate(ilistfile,
-		  "find . -xdev -path ./home -prune -o \\( \\( -type l -o -type f \\) -a -links +1 -print0 \\) | xargs -0 stat --format '%d %i '");
+
+      /* but first, try etch-workaround */
+      if (pc->debian_etch_workaround)
+	{
+	  fprintf(stderr, "W: Trying my backup method for working with Debian Etch chroots.\n");
+	  if (forkexeclp("chroot",
+			 "chroot",
+			 pc->buildplace,
+			 "cow-shell",
+			 "/bin/true",
+			 NULL))
+	    {
+	      /* I failed, what can I do? noooo */
+	      fprintf(stderr, "E: failed running Debian Etch compatibility backup plan, I'm in panic, eek.\n");
+	      return 1;
+	    }
+	}
+      else
+	{
+	  /*
+	     try the normal backup method
+
+	     It probably doesn't work but try it anyway.
+	   */
+	  fprintf(stderr, "W: cowdancer-ilistcreate failed to run within chroot, falling back to old method\n");
+	  ilistcreate(ilistfile,
+		      "find . -xdev -path ./home -prune -o \\( \\( -type l -o -type f \\) -a -links +1 -print0 \\) | xargs -0 stat --format '%d %i '");
+	}
     }
   chdir(prevdir);
   free(prevdir);
@@ -242,16 +312,16 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile_)
   pbuildercommandline[1]="build";
   PBUILDER_ADD_PARAM("--buildplace");
   PBUILDER_ADD_PARAM(pc->buildplace);
-  if (pc->mirror)
-    {
-      PBUILDER_ADD_PARAM("--mirror");
-      PBUILDER_ADD_PARAM(pc->mirror);
-    }
   if (pc->buildresult)
     {
       PBUILDER_ADD_PARAM("--buildresult");
       PBUILDER_ADD_PARAM(pc->buildresult);
     }
+  if (pc->debbuildopts)
+    {
+      PBUILDER_ADD_PARAM("--debbuildopts");
+      PBUILDER_ADD_PARAM(pc->debbuildopts);
+    }
   PBUILDER_ADD_PARAM("--no-targz");
   PBUILDER_ADD_PARAM("--internal-chrootexec");
   PBUILDER_ADD_PARAM(buf_chroot);
@@ -271,7 +341,7 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile_)
 int cpbuilder_create(const struct pbuilderconfig* pc)
 {
   int ret;
-  
+
   mkdir(pc->basepath,0777);
   printf(" -> Invoking pbuilder\n");
 
@@ -283,6 +353,13 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
       PBUILDER_ADD_PARAM("--mirror");
       PBUILDER_ADD_PARAM(pc->mirror);
     }
+  if (pc->arch)
+    {
+      PBUILDER_ADD_PARAM("--architecture");
+      PBUILDER_ADD_PARAM(pc->arch);
+    }
+  PBUILDER_ADD_PARAM("--distribution");
+  PBUILDER_ADD_PARAM(pc->distribution);
   PBUILDER_ADD_PARAM("--no-targz");
   PBUILDER_ADD_PARAM("--extrapackages");
   PBUILDER_ADD_PARAM("cowdancer");
@@ -290,9 +367,9 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
   ret=forkexecvp(pbuildercommandline);
 
   if (ret)
-    {	  
+    {
       printf("pbuilder create failed\n");
-      if (0!=forkexeclp("rm", "rm", "-rf", pc->basepath, NULL))
+      if (0!=rmrf(pc->basepath))
 	{
 	  fprintf(stderr, "Could not remove original tree\n");
 	}
@@ -312,12 +389,15 @@ int cpbuilder_internal_saveupdate(const struct pbuilderconfig* pc)
 {
   int ret=0;
   char* buf_tmpfile=NULL;
-  
+
   if (0>asprintf(&buf_tmpfile, "%s-%i-tmp", pc->buildplace, (int)getpid()))
     {
       /* outofmemory */
       return -1;
     }
+  printf(" -> Moving work directory [%s] to final location [%s] and cleaning up old copy\n",
+	 pc->buildplace, pc->basepath);
+
   if(-1==rename(pc->basepath, buf_tmpfile))
     {
       perror("rename");
@@ -330,7 +410,7 @@ int cpbuilder_internal_saveupdate(const struct pbuilderconfig* pc)
       ret=-1;
       goto out;
     }
-  if (0!=forkexeclp("rm", "rm", "-rf", buf_tmpfile, NULL))
+  if (0!=rmrf(buf_tmpfile))
     {
       printf("Could not remove original tree\n");
       ret=-1;
@@ -346,13 +426,13 @@ int cpbuilder_login(const struct pbuilderconfig* pc)
 {
   char *buf_chroot;
   int ret;
-  
+
   if (cpbuilder_internal_cowcopy(pc))
     {
       fprintf(stderr, "Failed cowcopy.\n");
       return 1;
     }
-  
+
   /* this is the option passed to internal-chrootexec */
   if (0>asprintf(&buf_chroot, "chroot %s cow-shell", pc->buildplace))
     {
@@ -389,7 +469,7 @@ int cpbuilder_login(const struct pbuilderconfig* pc)
   return ret;
 }
 
-/* 
+/*
 
 Mostly a copy of pbuilder login, executes a script.
 
@@ -399,13 +479,13 @@ int cpbuilder_execute(const struct pbuilderconfig* pc, char** av)
   char *buf_chroot;
   int ret;
   int i=0;
-  
+
   if (cpbuilder_internal_cowcopy(pc))
     {
       fprintf(stderr, "Failed cowcopy.\n");
       return 1;
     }
-  
+
   if (0>asprintf(&buf_chroot, "chroot %s cow-shell", pc->buildplace))
     {
       /* outofmemory */
@@ -444,11 +524,29 @@ int cpbuilder_execute(const struct pbuilderconfig* pc, char** av)
     }
   else
     printf("pbuilder execute aborted, not cleaning\n");
-  
+
   return ret;
 }
 
 
+/*
+   find matching parameter, and check if it's set.
+   @returns found=1, not found=0.
+ */
+int find_matching_param(const char* option)
+{
+  int i;
+  int found=0;
+  for (i=0; i<offset; ++i)
+    {
+      if (!strcmp(pbuildercommandline[i], option))
+	{
+	  found=1;
+	}
+    }
+  return found;
+}
+
 /**
    implement pbuilder update
 
@@ -458,8 +556,8 @@ int cpbuilder_update(const struct pbuilderconfig* pc)
 {
   char * buf_chroot;
   int ret;
-  
-  /* 
+
+  /*
      I want the following actions done:
      NEW=buildplace OLD=basepath
 
@@ -473,12 +571,18 @@ int cpbuilder_update(const struct pbuilderconfig* pc)
   */
 
   /* This simple process will not disrupt already-running pbuilder-cow session,
-     but it will not work with new pbuilder-cow sessions until this process finishes, 
+     but it will not work with new pbuilder-cow sessions until this process finishes,
      and any failure to pbuilder update will leave crap.
   */
-  if (0>asprintf(&buf_chroot, "chroot %s cow-shell", pc->buildplace))
+
+  if (0>asprintf(&buf_chroot,
+		 (pc->no_cowdancer_update)?"chroot %s":	/* try to not use cow-shell, when
+							   no-cowdancer-update option is used */
+		 "chroot %s cow-shell",
+		 pc->buildplace))
     {
-      /* outofmemory */
+      /* if outofmemory, die. */
+      fprintf(stderr, "Out of memory.\n");
       return -1;
     }
 
@@ -493,10 +597,15 @@ int cpbuilder_update(const struct pbuilderconfig* pc)
   pbuildercommandline[1]="update";
   PBUILDER_ADD_PARAM("--buildplace");
   PBUILDER_ADD_PARAM(pc->buildplace);
-  if (pc->mirror)
+  if (find_matching_param("--override-config"))
     {
-      PBUILDER_ADD_PARAM("--mirror");
-      PBUILDER_ADD_PARAM(pc->mirror);
+      if (pc->mirror)
+	{
+	  PBUILDER_ADD_PARAM("--mirror");
+	  PBUILDER_ADD_PARAM(pc->mirror);
+	}
+      PBUILDER_ADD_PARAM("--distribution");
+      PBUILDER_ADD_PARAM(pc->distribution);
     }
   PBUILDER_ADD_PARAM("--no-targz");
   PBUILDER_ADD_PARAM("--internal-chrootexec");
@@ -508,18 +617,25 @@ int cpbuilder_update(const struct pbuilderconfig* pc)
     {
       if (ret < 128)
 	{
-	  printf("pbuilder update failed\n");
-	  if (0!=forkexeclp("rm", "rm", "-rf", pc->buildplace, NULL))
+	  fprintf(stderr, "pbuilder update failed\n");
+
+	  if (!pc->no_cowdancer_update)
+	    {
+	      /* advise the user to try this option first */
+	      fprintf(stderr, "E: could not update with cowdancer, try --no-cowdancer-update option\n");
+	    }
+
+	  if (0!=rmrf(pc->buildplace))
 	    {
 	      fprintf(stderr, "Could not remove original tree\n");
 	    }
 	}
       else
-	  printf("pbuilder update aborted, not cleaning\n");
-      
+	printf("pbuilder update aborted, not cleaning\n");
+
       /* either way, I don't want to touch the original tree */
       goto out;
-      
+
     }
   printf(" -> removing %s working copy\n", ilist_PRGNAME);
   cpbuilder_internal_saveupdate(pc);
@@ -548,7 +664,7 @@ int cpbuilder_help(void)
   return 0;
 }
 
-int main(int ac, char** av)
+int app_main(int ac, char** av)
 {
   setenv("COWDANCER_REUSE","yes",1);
   return parse_parameter(ac, av, "cow");
