@@ -1,6 +1,6 @@
 /*BINFMTC:
  *  parameter handling for cpbuilder.
- *  Copyright (C) 2007-2008 Junichi Uekawa
+ *  Copyright (C) 2007-2009 Junichi Uekawa
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -18,6 +18,11 @@
  *
  */
 
+#ifndef __PARAMETER_H__
+#define __PARAMETER_H__
+
+#define MAX_CUSTOM_FILES 32
+
 typedef struct pbuilderconfig 
 {
   /* if you edit here, please add to parameter.c: dumpconfig */
@@ -25,6 +30,7 @@ typedef struct pbuilderconfig
   int mountdev;
   int mountdevpts;
   int save_after_login;
+  int debug;
   char* buildplace;		/* /var/cache/pbuilder/build/XXX.$$ */
   char* buildresult;		/* /var/cache/pbuilder/result/ */
   char* basepath;		/* /var/cache/pbuilder/cow */
@@ -32,7 +38,12 @@ typedef struct pbuilderconfig
   char* distribution;
   char* components;
   char* debbuildopts;
-  
+
+  /* files to be copied into the chroot, 
+     and copied out of the chroot */
+  char* inputfile[MAX_CUSTOM_FILES+1];
+  char* outputfile[MAX_CUSTOM_FILES+1];
+    
   /* cow-specific options */
   int no_cowdancer_update;		/* --no-cowdancer-update */
   int debian_etch_workaround;		/* --debian-etch-workaround */
@@ -57,7 +68,7 @@ typedef struct pbuilderconfig
 }pbuilderconfig;
 
 int load_config_file(const char* config, pbuilderconfig* pc);
-
+int size_of_ntarray(char ** buf);
 int forkexeclp (const char *path, const char *arg0, ...);
 int forkexecvp (char *const argv[]);
 int parse_parameter(int ac, char** av, const char* keyword);
@@ -88,3 +99,6 @@ PBUILDER_ADD_PARAM(NULL);
  {pbuildercommandline[offset]=NULL; fprintf(stderr, "pbuilder-command-line: Max command-line exceeded\n");}
 extern char* pbuildercommandline[MAXPBUILDERCOMMANDLINE];
 extern int offset;
+
+#endif
+
