@@ -1,6 +1,6 @@
-/*BINFMTC: parameter.c forkexec.c
+/*BINFMTC: parameter.c forkexec.c qemuipsanitize.c qemuarch.c file.c main.c
  *  qemubuilder: pbuilder with qemu
- *  Copyright (C) 2007-2008 Junichi Uekawa
+ *  Copyright (C) 2007-2009 Junichi Uekawa
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,20 +24,22 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
+#include <sys/socket.h>
 #include <string.h>
 #include <unistd.h>
 #include <getopt.h>
 #include <stdarg.h>
-#include <sys/types.h>
-#include <sys/socket.h>
 #include <assert.h>
 #include <termios.h>
 #include <time.h>
 #include <locale.h>
 #include "parameter.h"
+#include "qemuipsanitize.h"
+#include "qemuarch.h"
+#include "file.h"
 
-/* 
- * example exit codes: 
+/*
+ * example exit codes:
  *
  * END OF WORK EXIT CODE=1
  * END OF WORK EXIT CODE=0
@@ -45,148 +47,44 @@
  */
 const char* qemu_keyword="END OF WORK EXIT CODE=";
 
-/**
- * arch-specific routine; the console device to make
- */
-static const char* qemu_arch_serialdevice(const char* arch)
-{
-  if (!strcmp(arch, "arm") || 
-      !strcmp(arch, "armel"))
-    {
-      return "mknod dev/console c 204 64; ";
-    }
-  return "mknod dev/console c 4 64; ";
-}
-
-/**
- * arch-specific routine; disk device name to use
- */
-static const char* qemu_arch_diskdevice(const struct pbuilderconfig* pc)
-{
-  if (pc->arch_diskdevice)
-    return pc->arch_diskdevice;
-
-  if (!strcmp(pc->arch, "arm") || 
-      !strcmp(pc->arch, "armel"))
-    {
-      return "sd";
-    }
-  return "hd";
-}
-
-/**
- * arch-specific routine; qemu command to use.
- */
-static const char* qemu_arch_qemu(const char* arch)
-{
-  if (!strcmp(arch, "arm") || 
-      !strcmp(arch, "armel"))
-    {
-      return "qemu-system-arm";
-    }
-  if (!strcmp(arch, "mips"))
-    return "qemu-system-mips";
-  else if (!strcmp(arch, "mipsel"))
-    return "qemu-system-mipsel";
-  else if (!strcmp(arch, "i386") ||
-	   !strcmp(arch, "amd64"))
-    {
-      FILE*f=popen("dpkg --print-architecture", "r");
-      char*host_arch;
-      fscanf(f, "%as", &host_arch);
-      pclose(f);
-
-      /* special-case
-	 use kvm if possible
-	 use qemu if i386/i386
-	 use qemu-system-x86_64 otherwise
-       */
-      if(!strcmp(host_arch,arch) && !(system("which kvm")))
-	return "kvm";
-      else if((!strcmp(host_arch, "i386")) && (!strcmp(arch, "i386")))
-	return "qemu";
-      else 
-	return "qemu-system-x86_64";
-    }
-  else if (!strcmp(arch, "powerpc"))
-    return "qemu-system-ppc";
-  else
-    return NULL;
-}
-
-/**
- * arch-specific routine; the machine spec for this arch
- */
-static const char* qemu_arch_qemumachine(const char* arch)
-{
-  if (!strcmp(arch, "arm") || 
-      !strcmp(arch, "armel"))
-    return "versatilepb";
-  else if (!strcmp(arch, "i386") ||
-	   !strcmp(arch, "amd64"))
-    return "pc";
-  else if (!strcmp(arch, "mips")||
-	   !strcmp(arch, "mipsel"))
-    return "mips";
-  else if (!strcmp(arch, "powerpc"))
-    return "prep";
-  return NULL;
-}
-
-/**
- * arch-specific routine; the serial device
- */
-static const char* qemu_arch_tty(const char* arch)
-{
-
-  if (!strcmp(arch, "arm")||
-      !strcmp(arch, "armel"))
-    {
-      return "ttyAMA0";
-    }
-  return "ttyS0";
-}
-
-/** create a sparse ext3 block device suitable for 
+/** create a sparse ext3 block device suitable for
     loop-mount.
 
+    This code takes approx 7 seconds to run, should be cached?
+
    @returns -1 on error, 0 on success
  */
-static int create_ext3_block_device(const char* filename, int gigabyte)
+static int create_ext3_block_device(const char* filename,
+				    unsigned long int gigabyte)
 {
   int ret = 0;
   char *s = NULL;
   char *s2 = NULL;
 
   /* create 10GB sparse data */
-  if (0>asprintf(&s, "of=%s", filename))
-    {
-      /* out of memory */
-      return -1;
-    }
-  if (0>asprintf(&s2, "seek=%i", gigabyte * 1024))
-    {
-      /* out of memory */
-      return -1;
-    }
-  ret=forkexeclp("dd", "dd", "if=/dev/zero", s,
-		 "bs=1M",
-		 s2, "count=1", NULL);
-  if (ret)
+  if (create_sparse_file(filename,
+			 gigabyte*1UL<<30UL))
     {
       ret=-1;
       goto out;
     }
 
-  free(s); s=NULL;
-
-  if (0>asprintf(&s, "mke2fs -F -j -O sparse_super %s", filename))
+  if ((ret=forkexeclp("mke2fs", "mke2fs",
+		      "-q",
+		      "-F",
+		      "-j",
+		      "-m1",
+		      "-O", "sparse_super",
+		      filename, NULL)))
     {
-      /* out of memory */
       ret=-1;
       goto out;
     }
-  if ((ret=system(s)))
+
+  if ((ret=forkexeclp("tune2fs", "tune2fs",
+		      "-c", "0",
+		      "-i", "0",
+		      filename, NULL)))
     {
       ret=-1;
     }
@@ -219,94 +117,73 @@ static int loop_umount(const char* device)
   return ret;
 }
 
-/** create a script here. 
-    
-@returns -1 on failure, 0 on success.
+/** create a script file.
+
+@returns NULL on failure, FILE* on success
 */
-static int create_script(const char* mountpoint, const char* relative_path, const char* scriptcontent)
+static FILE* create_script(const char* mountpoint, const char* relative_path)
 {
   char *s = NULL;
-  FILE* f = NULL;
-  int ret=-1;
-  
+  FILE *f = NULL;
+  FILE *ret = NULL;
+
   asprintf(&s, "%s/%s", mountpoint, relative_path);
   if(!(f=fopen(s, "w")))
     goto fail;
-  if(EOF==fputs(scriptcontent, f))
-    goto fail;
-  if(EOF==fclose(f))
-    goto fail;
   if(chmod(s, 0700))
-    goto fail;
-  ret=0;
+    {
+      fclose(f);
+      goto fail;
+    }
+  ret=f;
  fail:
   free(s);
   return ret;
 }
 
-/**
-   copy file contents to temporary filesystem, so that it can be used inside qemu.
- */
-static int copy_file_internal(const char* orig, const char*dest)
-{
-  const int buffer_size=4096;
-  char *buf = malloc (buffer_size);
-  int ret=-1;
-  FILE* fin;
-  FILE* fout;
-  size_t count;
-
-  fin=fopen(orig, "r");
-  fout=fopen(dest, "w");
-  if (!fin)
-    {
-      goto out;
-    }
-  if (!fout)
-    {
-      goto out;
-    }
-  while((count=fread(buf, 1, buffer_size, fin)))
-    {
-      fwrite(buf, 1, count, fout);
-    }
-  if(ferror(fin))
-    {
-      /* TODO: handle error */
-      goto out;
-    }
-  fclose(fin);
-  fclose(fout);
-  ret=0;
- out:
-  if (buf) free (buf);
-  
-  return ret;
-}
-
 /* minimally fix terminal I/O */
 static void fix_terminal(void)
 {
-  struct termios t; 
+  struct termios t;
 
-  tcgetattr(1, &t); 
-  t.c_lflag |= ECHO;
-  tcsetattr(1, TCSANOW, &t); 
+  if (isatty(1))
+    {
+      tcgetattr(1, &t);
+      t.c_lflag |= ECHO;
+      tcsetattr(1, TCSANOW, &t);
+    }
 }
 
-static int copy_file_contents_to_temp(const char* orig, const char*temppath, const char* filename)
+static int copy_file_contents_through_temp(FILE* f,
+					   const char* orig,
+					   const char* temppath,
+					   const char* targetdir)
 {
-  char* s;
+  char* tempname;
   int ret;
+  char* file_basename = basename(orig);
+  asprintf(&tempname, "%s/%s", temppath, file_basename);
+  ret=copy_file(orig, tempname);
+  if (ret == -1)
+    {
+      fprintf(f, "E: Copy file error in %s to %s\n",
+	      orig, tempname);
+    }
 
-  asprintf(&s, "%s/%s", temppath, filename);
-  ret=copy_file_internal(orig, s);
-  free(s);
+  free(tempname);
+
+  fprintf(f,
+	  "echo \"I: copying %s/%s from temporary location\"\n"
+	  "cp $PBUILDER_MOUNTPOINT/%s %s/%s || echo \"E: Copy failed\"\n",
+	  targetdir, file_basename,
+	  file_basename,
+	  targetdir, file_basename);
   return ret;
 }
 
 /**
-   run qemu until exit signal is given.
+   run qemu until exit signal is received from within QEMU via serial
+   console.
 
    exit code:
    -1: error
@@ -321,20 +198,21 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
   const int buffer_size=4096;
   char* buf=malloc(buffer_size);
   size_t count;
-  
+
   if (-1==socketpair(AF_UNIX, SOCK_STREAM,0,sp))
     {
       /* error handle? */
       return -1;
     }
-  
+
+  fflush(NULL);
   if ((child=fork()))
     {
       /* this is parent process */
 
       close(sp[1]);
       close(0);
-      
+
       FD_ZERO(&readfds);
       while (1)
 	{
@@ -346,23 +224,23 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 		  void* matchptr;
 
 		  /* data available from qemu */
-		  
+
 		  /* sleep a bit to let it buffer-up a bit more. */
 		  usleep(100000);
-		  
+
 		  count=read(sp[0],buf,buffer_size);
-		  
+
 		  /* this won't work sometimes, but this is a good best-effort thing. */
-		  if ((matchptr=memmem(buf, count, 
+		  if ((matchptr=memmem(buf, count,
 				       qemu_keyword, strlen(qemu_keyword)))!=0)
 		    {
 		      int status;
 
 		      exit_code = atoi(matchptr + strlen(qemu_keyword));
-		      printf("\n  -> received termination message from inside qemu with exit-code %i, killing child process (qemu:%i)\n", 
+		      printf("\n  -> received termination message from inside qemu with exit-code %i, killing child process (qemu:%i)\n",
 			     exit_code,
 			     child);
-		      
+
 		      assert(child != 0);assert(child > 0);
 
 		      if (!kill(child, SIGTERM))
@@ -388,55 +266,66 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
   else if(child == 0)
     {
       /* this is the child process */
-      const char* qemu = qemu_arch_qemu(pc->arch); 
+      const char* qemu = qemu_arch_qemu(pc->arch);
       const char* machine = qemu_arch_qemumachine(pc->arch);
       char* append_command;
       const char* kernel_image = pc->kernel_image;
       const char* initrd = pc->initrd;
       char* mem;
+      int argc = 0;
+      const int MAX_ARGS = 30;
+      char *argv[MAX_ARGS];
+      int i;
 
       if (qemu == NULL || machine == NULL) {
-	fprintf(stderr, "Your arch %s does not seem to be supported\n", pc->arch);
+	fprintf(stderr, "Your architecture %s does not seem to be supported\n", pc->arch);
 	exit(1);
       }
 
       asprintf(&mem, "%i", pc->memory_megs);
-      
+
       asprintf(&append_command,
-	       "root=/dev/%sa init=/pbuilder-run console=%s",
+	       "root=/dev/%sa quiet init=/pbuilder-run console=%s",
 	       qemu_arch_diskdevice(pc),
 	       qemu_arch_tty(pc->arch));
-            
+
       dup2(sp[1],1);
       dup2(sp[1],2);
       close(sp[0]);
-      
-      if (initrd)
+
+      argv[argc++]=strdupa(qemu);
+      argv[argc++]="-nographic";
+      argv[argc++]="-M";
+      argv[argc++]=strdupa(machine);
+      argv[argc++]="-m";
+      argv[argc++]=mem;
+      argv[argc++]="-kernel";
+      argv[argc++]=strdupa(kernel_image);
+      if (initrd && strcmp(initrd, ""))
 	{
-	  execlp(qemu, qemu, 
-		 "-nographic",
-		 "-M", machine, 
-		 "-m", mem,
-		 "-kernel", kernel_image, 
-		 "-initrd", initrd, 
-		 "-hda", hda,
-		 "-hdb", hdb, 
-		 "-append", append_command, 
-		 "-serial", "stdio", NULL);
+	  argv[argc++]="-initrd";
+	  argv[argc++]=strdupa(initrd);
 	}
-      else
+      argv[argc++]="-hda";
+      argv[argc++]=strdupa(hda);
+      argv[argc++]="-hdb";
+      argv[argc++]=strdupa(hdb);
+      argv[argc++]="-append";
+      argv[argc++]=append_command;
+      argv[argc++]="-serial";
+      argv[argc++]="stdio";
+      argv[argc]=NULL;
+      assert(argc < MAX_ARGS);
+
+      printf("  forking qemu: ");
+      for (i=0; i<argc; ++i)
 	{
-	  execlp(qemu, qemu, 
-		 "-nographic",
-		 "-M", machine, 
-		 "-m", mem,
-		 "-kernel", kernel_image, 
-		 "-hda", hda,
-		 "-hdb", hdb, 
-		 "-append", append_command, 
-		 "-serial", "stdio", NULL);
+	  printf("%s ", argv[i]);
 	}
-      perror("execlp");
+      printf("\n");
+
+      execvp(argv[0], argv);
+      perror("fork_qemu");
       exit (1);
     }
   else
@@ -451,16 +340,39 @@ static int fork_qemu(const char* hda, const char* hdb, const struct pbuilderconf
 
 static int do_fsck(const char* devfile)
 {
+  /* force-running this fsck isn't a good idea; let it fail.
+     If it's mounted by someone else, I don't want to touch it,
+     and chroots can be re-created any time, right?
+   */
   return forkexeclp("/sbin/fsck",
 		    "/sbin/fsck",
-		    devfile, 
+		    devfile,
 		    NULL);
 }
 
+/*
+   get the current time string which can be used in date command to
+   set time inside the chroot.
+ */
+static char* get_current_time_string(void)
+{
+  char* locsave, *timestring;
+  time_t currenttime;
+
+  /* save/set/restore locale settings to get current time in POSIX format */
+  locsave = setlocale(LC_TIME, NULL);
+  (void) setlocale(LC_TIME, "POSIX");
+  currenttime=time(NULL);
+  timestring=asctime(gmtime(&currenttime));
+  (void) setlocale(LC_TIME, locsave);
+  return timestring;
+}
+
+
 /**
  * Invoke qemu, and run the second-stage script within QEMU.
  *
- * 
+ * hostcommand1 is used from build and login and exeute
  */
 static int run_second_stage_script
 (
@@ -477,10 +389,11 @@ static int run_second_stage_script
   char* script=NULL;
   char* workblockdevicepath=NULL;
   char* cowdevpath=NULL;
-  char* locsave, *timestring;
-  time_t currenttime;
+  char* timestring;
   int ret=1;
-  
+  FILE* f;
+  int i;
+
   if (mkdir(pc->buildplace,0777))
     {
       /* could not create the buildplace here. */
@@ -490,60 +403,68 @@ static int run_second_stage_script
 
   do_fsck(pc->basepath);
 
-  /* save/set/restore locale settings to get current time in POSIX format */
-  locsave = setlocale(LC_TIME, NULL);
-  (void) setlocale(LC_TIME, "POSIX");
-  currenttime=time(NULL); 
-  timestring=asctime(gmtime(&currenttime));
-  (void) setlocale(LC_TIME, locsave);
+  timestring=get_current_time_string();
 
   asprintf(&workblockdevicepath, "%s.dev", pc->buildplace);
   ret=create_ext3_block_device(workblockdevicepath, 1);
   loop_mount(workblockdevicepath, pc->buildplace);
 
-  asprintf(&script, 
-	   "#!/bin/bash\n"
-	   "echo \n"
-	   "echo ' -> qemu-pbuilder second-stage' \n"
-	   //TODO: copy hook scripts
-	   //"mount -n /proc /proc -t proc\n" // this is done in first stage.
-	   "echo '  -> setting time to %s' \n"
-	   "date --set=\"%s\"\n"
-	   "echo '  -> configuring network' \n"
-	   "ifconfig -a\n"
-	   "export IFNAME=`/sbin/ifconfig -a | grep eth | head -n1 | awk '{print $1}'`\n"
-	   "dhclient $IFNAME\n"
-	   "cp $PBUILDER_MOUNTPOINT/hosts /etc/hosts\n"
-	   "cp $PBUILDER_MOUNTPOINT/resolv.conf /etc/resolv.conf\n"
-	   "cp $PBUILDER_MOUNTPOINT/hostname /etc/hostname\n"
-	   "hostname pbuilder-$(cat /etc/hostname)\n"
-	   //TODO: run G hook
-	   "%s\n"
-	   "apt-get clean || true\n"
-	   "sync\n"
-	   "sync\n"
-	   "while sleep 3s; do\n"
-	   "  echo ' -> qemu-pbuilder %s0'\n"
-	   "done\n",
-	   timestring,
-	   timestring,
-	   commandline, 
-	   qemu_keyword
-	   );
+  f = create_script(pc->buildplace, "pbuilder-run");
+  fprintf(f,
+	  "#!/bin/bash\n"
 
-  create_script(pc->buildplace, "pbuilder-run",
-		script);
+	  /* define function to terminate qemu */
+	  "function exit_from_qemu() {\n"
+	  "sync\n"
+	  "sync\n"
+	  "sleep 1s\n"		/* sleep before sending dying message */
+	  "echo ' -> qemu-pbuilder %s$1'\n"
+	  "sleep 1s\n"
+	  "halt -f -p\n"	/* just halt myself if possible */
+	  "}\n",
+	  qemu_keyword);
 
-  /* TODO: copy /etc/hosts etc. to inside chroot, or it won't know the hosts info. */
-  copy_file_contents_to_temp("/etc/hosts", pc->buildplace, "hosts");
-  copy_file_contents_to_temp("/etc/hostname", pc->buildplace, "hostname");
-  copy_file_contents_to_temp("/etc/resolv.conf", pc->buildplace, "resolv.conf");
+  fprintf(f,
+	  /* main code */
+	  "echo \n"
+	  "echo ' -> qemu-pbuilder second-stage' \n"
+	  //TODO: copy hook scripts
+	  //"mount -n /proc /proc -t proc\n" // this is done in first stage.
+	  "echo '  -> setting time to %s' \n"
+	  "date --set=\"%s\"\n"
+	  "echo '  -> configuring network' \n"
+	  "ifconfig -a\n"
+	  "export IFNAME=`/sbin/ifconfig -a | grep eth | head -n1 | awk '{print $1}'`\n"
+	  "dhclient $IFNAME\n"
+	  "mkdir -p /tmp/buildd\n"
+	  "$PBUILDER_MOUNTPOINT/run-copyfiles\n"
+	  "hostname pbuilder-$(cat /etc/hostname)\n"
+	  //TODO: run G hook
+	  "%s\n"
+	  //TODO: I can mount /var/cache/apt/archives from some scratch space to not need this:
+	  "apt-get clean || true\n"
+	  "exit_from_qemu 0\n",
+	  timestring,
+	  timestring,
+	  commandline);
+  fclose(f);
+
+  /* copy files script */
+  f = create_script(pc->buildplace, "run-copyfiles");
+  copy_file_contents_through_temp(f, "/etc/hosts", pc->buildplace, "/etc");
+  copy_file_contents_through_temp(f, "/etc/hostname", pc->buildplace, "/etc");
+  /* copy inputfile */
+  for (i=0; pc->inputfile[i]; ++i)
+    {
+      copy_file_contents_through_temp(f, pc->inputfile[i], pc->buildplace, "/tmp/buildd");
+    }
+  fclose(f);
 
   /* do I not need to copy /etc/pbuilderrc, and ~/.pbuilderrc to inside chroot? */
   /* TODO: hooks probably need copying here. */
   /* TODO: recover aptcache */
 
-  if(hostcommand1) 
+  if(hostcommand1)
     {
       printf("running host command: %s\n", hostcommand1);
       system(hostcommand1);
@@ -552,17 +473,15 @@ static int run_second_stage_script
   loop_umount(workblockdevicepath);
 
   asprintf(&cowdevpath, "%s.cowdev", pc->buildplace);
-  ret=forkexeclp("qemu-img", "qemu-img", 
+  ret=forkexeclp("qemu-img", "qemu-img",
 		 "create",
 		 "-f",
-		 "qcow", 
-		 "-b", 
-		 pc->basepath, 
-		 cowdevpath, 
+		 "qcow",
+		 "-b",
+		 pc->basepath,
+		 cowdevpath,
 		 NULL);
 
-  // this will have wrong time within qemu, for ARM
-  // how to workaround?
   fork_qemu(cowdevpath, workblockdevicepath, pc);
   /* this will always return 0. */
 
@@ -570,16 +489,16 @@ static int run_second_stage_script
   if (save_result)
     {
       printf(" -> commit change to qemu image\n");
-      ret=forkexeclp("qemu-img", "qemu-img", 
+      ret=forkexeclp("qemu-img", "qemu-img",
 		     "commit",
-		     cowdevpath, 
-		     NULL);  
+		     cowdevpath,
+		     NULL);
     }
 
   /* after-run */
   loop_mount(workblockdevicepath, pc->buildplace);
   printf(" -> running post-run process\n");
-  if(hostcommand2) 
+  if(hostcommand2)
     {
       printf("running host command: %s\n", hostcommand2);
       system(hostcommand2);
@@ -598,7 +517,7 @@ static int run_second_stage_script
   return ret;
 }
 
-/* 
+/*
    @return shell command to copy the dsc file.
  */
 static char* copy_dscfile(const char* dscfile_, const char* destdir)
@@ -617,9 +536,11 @@ static char* copy_dscfile(const char* dscfile_, const char* destdir)
   int filelist=0;
 
   origdir=strdup(dscfile);
-  (*(strrchr(origdir,'/')))=0;	/* if this isn't working, I'm dead */
-    
-  fprintf(fmem, "cp %s %s/\n", 
+
+  assert(strrchr(origdir,'/') != 0);
+  (*(strrchr(origdir,'/')))=0;
+
+  fprintf(fmem, "cp %s %s/\n",
 	  dscfile, destdir);
 
   while (getline(&buf,&bufsiz,f)>0)
@@ -634,19 +555,19 @@ static char* copy_dscfile(const char* dscfile_, const char* destdir)
 	    filelist=0;
 	  else
 	    {
-	      fprintf(fmem, "cp %s/%s %s/\n", 
+	      fprintf(fmem, "cp %s/%s %s/\n",
 		      origdir, filename, destdir);
 	      assert(filename);
 	      free(filename);
 	    }
-	}  
+	}
       if (!(buf[0]==' ')&&
 	  !strcmp(buf,"Files: "))
 	{
 	  filelist=1;
 	}
-      
     }
+
   ret=0;
   assert(fmem);
   assert(f);
@@ -659,17 +580,25 @@ static char* copy_dscfile(const char* dscfile_, const char* destdir)
   return ret?NULL:memstr;
 }
 
+/*
+   return 0 on success, nonzero on failure.
+
+   variable ret holds the state.
+ */
 int cpbuilder_create(const struct pbuilderconfig* pc)
 {
-  int ret;
-  char* s=NULL;  
+  int ret=0;
+  char* s=NULL;  		/* generic asprintf buffer */
   char* workblockdevicepath=NULL;
+  FILE* f;
+  char* t;
+  char* timestring;
+
+  /* remove existing file; it can be old qemu image, or a directory if
+     it didn't exist before. */
+  unlink(pc->basepath);
+  rmdir(pc->basepath);
 
-  if((ret=unlink(pc->basepath)))
-    {
-      //ignore this
-    }
-  
   /* 3GB should be enough to install any Debian system; hopefully */
   ret=create_ext3_block_device(pc->basepath, 3);
 
@@ -677,10 +606,11 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
     {
       goto out;
     }
-  
+
   if (mkdir(pc->buildplace,0777))
     {
       /* could not create the buildplace here. */
+      ret=1;
       perror("mkdir");
       goto out;
     }
@@ -692,63 +622,51 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
     }
 
   printf(" -> Invoking debootstrap\n");
-  ret=forkexeclp("debootstrap", "debootstrap", 
-		 "--arch", 
-		 pc->arch, 
+  ret=forkexeclp("debootstrap", "debootstrap",
+		 "--arch",
+		 pc->arch,
 		 "--foreign",
 		 pc->distribution,
-		 pc->buildplace, 
-		 pc->mirror, 
+		 pc->buildplace,
+		 pc->mirror,
 		 NULL);
   if (ret)
     {
+      fprintf(stderr, "debootstrap failed with %i\n",
+	      ret);
       goto umount_out;
     }
 
   /* arch-dependent code here.
      create required device files.
-   */
 
-  /* 
-     ttyAMA0 is probably ARM-specific 
+     ttyAMA0 is probably ARM-specific
      others are probably linux-portable as documented in linux/Documentation/devices.txt
-     other OSes will require different, but hey, they probably don't even boot from ext3, 
+     other OSes will require different, but hey, they probably don't even boot from ext3,
      we'll need think of other ways to work with them.
+
    */
-  asprintf(&s, 
-	   "cd %s; "
-	   "%s"
-	   "mknod dev/ttyS0 c 4 64; "
-	   "mknod dev/ttyAMA0 c 204 64; "
-	   "mknod dev/sda b 8 0; mknod dev/sdb b 8 16; "
-	   "mknod dev/hda b 3 0; mknod dev/hdb b 3 64; ", 
-	   pc->buildplace,
-	   qemu_arch_serialdevice(pc->arch)
-	   );
-  if ((ret=system(s)))
-    {
-      goto umount_out;
-    }
-  free(s);
-  
-  asprintf(&s,
-	   "#!/bin/bash\n"
-	   "echo \n"
-	   "echo ' -> qemu-pbuilder first-stage' \n"
-	   "export PATH='/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin'\n"
-	   "mount -n /proc /proc -t proc\n"
-	   "mount -n -o rw,remount /\n"
-	   "cp /proc/mounts /etc/mtab\n"
-	   "export PBUILDER_MOUNTPOINT=/var/cache/pbuilder/pbuilder-mnt\n"
-	   "mkdir -p $PBUILDER_MOUNTPOINT\n"
-	   "mount -n -t ext3 /dev/%sb $PBUILDER_MOUNTPOINT \n"
-	   "$PBUILDER_MOUNTPOINT/pbuilder-run \n",
-	   qemu_arch_diskdevice(pc)
-	   );
-  
-  create_script(pc->buildplace, 
-		"pbuilder-run",
-		s);
+  printf(" -> Doing architecture-specific /dev population\n");
+
+  qemu_create_arch_devices(pc->buildplace, pc->arch);
+
+  f = create_script(pc->buildplace, "pbuilder-run");
+  fprintf(f,
+	  "#!/bin/bash\n"
+	  "echo \n"
+	  "echo ' -> qemu-pbuilder first-stage' \n"
+	  "export PATH='/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin'\n"
+	  "mount -n /proc /proc -t proc\n"
+	  "mount -n -o rw,remount /\n"
+	  "cp /proc/mounts /etc/mtab\n"
+	  "export PBUILDER_MOUNTPOINT=/var/cache/pbuilder/pbuilder-mnt\n"
+	  "mkdir -p $PBUILDER_MOUNTPOINT\n"
+	  "mount -n -t ext3 /dev/%sb $PBUILDER_MOUNTPOINT \n"
+	  "$PBUILDER_MOUNTPOINT/pbuilder-run \n",
+	  qemu_arch_diskdevice(pc)
+	  );
+  fclose(f);
+
   free(s); s=0;
 
   ret=loop_umount(pc->buildplace);
@@ -759,57 +677,78 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
 
   loop_mount(workblockdevicepath, pc->buildplace);
 
-  /* this is specific to my configuration, fix it later. */
-  asprintf(&s,
-	   "#!/bin/bash\n"
-	   "export RET=0\n"
-	   "echo \n"
-	   "echo ' -> qemu-pbuilder second-stage' \n"
-	   "/debootstrap/debootstrap --second-stage\n"
-	   "echo deb %s %s %s > /etc/apt/sources.list \n"
-	   "echo 'APT::Install-Recommends \"false\"; ' > /etc/apt/apt.conf.d/15pbuilder\n"
-	   //TODO: copy hook scripts
-	   "mount -n /proc /proc -t proc\n"
-	   "dhclient eth0\n"
-	   "cp $PBUILDER_MOUNTPOINT/hosts /etc/hosts\n"
-	   "cp $PBUILDER_MOUNTPOINT/resolv.conf /etc/resolv.conf\n"
-	   "cp $PBUILDER_MOUNTPOINT/hostname /etc/hostname\n"
-	   "hostname pbuilder-$(cat /etc/hostname)\n"
-	   //TODO: installaptlines
-	   //"echo 'deb http://192.168.1.26/debian/ sid main ' > /etc/apt/sources.list\n"
-	   //TODO: run G hook
-	   "apt-get update\n"
-	   //TODO: "dpkg --purge $REMOVEPACKAGES\n"
-	   //recover aptcache
-	   "apt-get -y --force-yes -o DPkg::Options::=--force-confnew dist-upgrade\n"
-	   "apt-get install --force-yes -y build-essential dpkg-dev apt aptitude pbuilder\n"
-	   //TODO: EXTRAPACKAGES handling
-	   //save aptcache
-	   //optionally autoclean aptcache
-	   //run E hook
-	   "apt-get clean || true\n"
-	   "export RET=0"
-	   ":EXIT"
-	   "sync\n"
-	   "sync\n"
-	   "while sleep 3s; do\n"
-	   "  echo \" -> qemu-pbuilder %s$RET\"\n"
-	   "done\n"
-	   "bash\n",
-	   pc->mirror, 
-	   pc->distribution,
-	   pc->components,
-	   qemu_keyword);
-  create_script(pc->buildplace,
-		"pbuilder-run",
-		s);
+  timestring=get_current_time_string();
+
+  f = create_script(pc->buildplace, "pbuilder-run");
+  fprintf(f,
+	  "#!/bin/bash\n"
+	  /* define function to terminate qemu */
+	  "function exit_from_qemu() {\n"
+	  "%s\n"
+	  "sync\n"
+	  "sync\n"
+	  "sleep 1s\n"		/* sleep before sending dying message */
+	  "echo ' -> qemu-pbuilder %s$1'\n"
+	  "sleep 1s\n"
+	  "halt -f -p\n"	/* just halt myself if possible */
+	  "}\n",
+	  pc->debug?"echo \"Debug shell\"; /bin/bash":"",
+	  qemu_keyword);
+
+  fprintf(f,
+	  /* start of main code */
+	  "export RET=0\n"
+	  "echo \n"
+	  "echo ' -> qemu-pbuilder second-stage' \n"
+	  "echo '  -> setting time to %s' \n"
+	  "date --set=\"%s\"\n"
+	  "echo '  -> Running debootstrap second-stage script' \n"
+	  "touch /etc/udev/disabled\n" // work-around for #520742
+	  "/debootstrap/debootstrap --second-stage || ( "
+	  "  echo dumping debootstrap log\n"
+	  "  cat /debootstrap/debootstrap.log\n"
+	  "  exit_from_qemu\n"
+	  "\n )\n"
+	  "rm /etc/udev/disabled\n" // work-around for #520742
+	  "echo deb %s %s %s > /etc/apt/sources.list \n"
+	  "echo 'APT::Install-Recommends \"false\"; ' > /etc/apt/apt.conf.d/15pbuilder\n"
+	  //TODO: copy hook scripts
+	  "mount -n proc /proc -t proc\n"
+	  "mount -n sysfs /sys -t sysfs\n"
+	  "mkdir /dev/pts\n"
+	  "mount -n devpts /dev/pts -t devpts\n"
+	  "dhclient eth0\n"
+	  "$PBUILDER_MOUNTPOINT/run-copyfiles\n"
+	  "hostname pbuilder-$(cat /etc/hostname)\n"
+	  //TODO: installaptlines
+	  //"echo 'deb http://192.168.1.26/debian/ sid main ' > /etc/apt/sources.list\n"
+	  //TODO: run G hook
+	  "apt-get update || exit_from_qemu 1\n"
+	  //TODO: "dpkg --purge $REMOVEPACKAGES\n"
+	  //recover aptcache
+	  "apt-get -y --force-yes -o DPkg::Options::=--force-confnew dist-upgrade || exit_from_qemu 1\n"
+	  "apt-get install --force-yes -y build-essential dpkg-dev apt aptitude pbuilder || exit_from_qemu 1\n"
+	  //TODO: EXTRAPACKAGES handling
+	  //save aptcache
+	  //optionally autoclean aptcache
+	  //run E hook
+	  //TODO: I can mount /var/cache/apt/archives from some scratch space to not need this:
+	  "apt-get clean || true\n"
+	  "exit_from_qemu $RET\n"
+	  "bash\n",
+	  timestring,
+	  timestring,
+	  t=sanitize_mirror(pc->mirror), pc->distribution, pc->components);
+  fclose(f);
+  free(t);
 
   /* TODO: can I do 'date --set' from output of 'LC_ALL=C date' */
 
-  /* TODO: copy /etc/hosts etc. to inside chroot, or it won't know the hosts info. */
-  copy_file_contents_to_temp("/etc/hosts", pc->buildplace, "hosts");
-  copy_file_contents_to_temp("/etc/hostname", pc->buildplace, "hostname");
-  copy_file_contents_to_temp("/etc/resolv.conf", pc->buildplace, "resolv.conf");
+  /* copy files script */
+  f = create_script(pc->buildplace, "run-copyfiles");
+  copy_file_contents_through_temp(f, "/etc/hosts", pc->buildplace, "/etc");
+  copy_file_contents_through_temp(f, "/etc/hostname", pc->buildplace, "/etc");
+  fclose(f);
 
   /* do I not need to copy /etc/pbuilderrc, and ~/.pbuilderrc to inside chroot? */
   /* TODO: hooks probably need copying here. */
@@ -817,14 +756,14 @@ int cpbuilder_create(const struct pbuilderconfig* pc)
 
   loop_umount(workblockdevicepath);
   rmdir(pc->buildplace);
-  
+
   // this will have wrong time. how to workaround?
   ret=fork_qemu(pc->basepath, workblockdevicepath, pc);
-  
+
   unlink(workblockdevicepath);
-  
+
  out:
-  if(workblockdevicepath) 
+  if(workblockdevicepath)
     {
       free(workblockdevicepath);
     }
@@ -848,7 +787,7 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
   char* commandline=NULL;
 
   const char* buildopt="--binary-all"; /* TODO: add --binary-arch option */
-  
+
   hoststr=copy_dscfile(dscfile, pc->buildplace);
 
   asprintf(&commandline,
@@ -865,13 +804,13 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
   /* Obscure assumption!: assume _ is significant for package name and
      no other file will have _. */
 
-  asprintf(&hoststr2, 
+  asprintf(&hoststr2,
 	   "cp -p \"%s\"/*_* \"%s\" 2>/dev/null || true",
 	   pc->buildplace, pc->buildresult);
-  
+
   ret=run_second_stage_script
     (0,
-     commandline, pc, 
+     commandline, pc,
      hoststr,
      hoststr2);
 
@@ -883,10 +822,14 @@ int cpbuilder_build(const struct pbuilderconfig* pc, const char* dscfile)
 
 int cpbuilder_login(const struct pbuilderconfig* pc)
 {
-  return run_second_stage_script(pc->save_after_login, "bash", pc, NULL, NULL);
+  return run_second_stage_script(pc->save_after_login,
+				 "bash",
+				 pc,
+				 NULL,
+				 NULL);
 }
 
-/* 
+/*
 
 Mostly a copy of pbuilder login, executes a script.
 
@@ -896,11 +839,11 @@ int cpbuilder_execute(const struct pbuilderconfig* pc, char** av)
   char* hostcommand;
   char* runcommandline;
   int ret;
-  
+
   asprintf(&hostcommand, "cp %s %s/run\n", av[0], pc->buildplace);
   /* TODO: add options too */
   asprintf(&runcommandline, "sh $PBUILDER_MOUNTPOINT/run");
-  ret=run_second_stage_script(pc->save_after_login, runcommandline, pc, 
+  ret=run_second_stage_script(pc->save_after_login, runcommandline, pc,
 			      hostcommand, NULL);
   free(hostcommand);
   free(runcommandline);
@@ -914,18 +857,21 @@ int cpbuilder_execute(const struct pbuilderconfig* pc, char** av)
  */
 int cpbuilder_update(const struct pbuilderconfig* pc)
 {
-  /* TODO: --override-config */
+  /* TODO: --override-config support.
+     There is no way to change distribution in this code-path...
+   */
   return run_second_stage_script
-    (1,	   
+    (1,
      //TODO: installaptlines if required.
      //TODO: "dpkg --purge $REMOVEPACKAGES\n"
-     "apt-get update\n"
+     //TODO: add error code handling.
+     "apt-get update -o Acquire::PDiffs=false\n"
      "apt-get -y --force-yes -o DPkg::Options::=--force-confnew dist-upgrade\n"
      "apt-get install --force-yes -y build-essential dpkg-dev apt aptitude pbuilder\n"
      //TODO: EXTRAPACKAGES handling
      //optionally autoclean aptcache
      //run E hook
-     , pc, 
+     , pc,
      NULL, NULL);
 }
 
@@ -949,7 +895,7 @@ int cpbuilder_help(void)
   return 0;
 }
 
-int main(int ac, char** av)
+int app_main(int ac, char** av)
 {
   return parse_parameter(ac, av, "qemu");
 }
